# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2019 by XESS Corp.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
GUI for finding/displaying footprints.
"""

from __future__ import print_function

import os
import string

import pykicad.module as pym
import wx

from skidl import (
    KICAD,
    footprint_search_paths,
    rmv_quotes,
    search_footprints_iter,
    skidl_cfg,
)
from .common import *
from .footprint_painter import FootprintPainter

APP_TITLE = "SKiDL Footprint Search"

APP_EXIT = 1
SHOW_HELP = 3
SHOW_ABOUT = 4
SEARCH_PATH = 5
SEARCH_FOOTPRINTS = 6
COPY_FOOTPRINTS = 7
PAINT_ACTUAL_SIZE_CKBX_ID = 8

footprint_search_text_id = 0


class AppFrame(wx.Frame):
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

        self.panel = FootprintSearchPanel(self)
        box = wx.BoxSizer(wx.VERTICAL)
        box.Add(self.panel, proportion=1, flag=wx.ALL | wx.EXPAND, border=SPACING)
        self.SetSizer(box)

        # Keep border same color as background of panel.
        self.SetBackgroundColour(self.panel.GetBackgroundColour())

        self.InitMenus()

        self.SetTitle(APP_TITLE)
        self.Center()
        self.Show(True)
        self.Fit()

    def InitMenus(self):

        # Top menu.
        menuBar = wx.MenuBar()

        # File submenu containing quit button.
        fileMenu = wx.Menu()
        menuBar.Append(fileMenu, "&File")

        quitMenuItem = wx.MenuItem(fileMenu, APP_EXIT, "Quit\tCtrl+Q")
        fileMenu.Append(quitMenuItem)
        self.Bind(wx.EVT_MENU, self.OnQuit, id=APP_EXIT)

        # Search submenu containing search and copy buttons.
        srchMenu = wx.Menu()
        menuBar.Append(srchMenu, "&Search")

        srchPathItem = wx.MenuItem(srchMenu, SEARCH_PATH, "Set search path...\tCtrl+P")
        srchMenu.Append(srchPathItem)
        self.Bind(wx.EVT_MENU, self.OnSearchPath, id=SEARCH_PATH)

        srchMenuItem = wx.MenuItem(srchMenu, SEARCH_FOOTPRINTS, "Search\tCtrl+F")
        srchMenu.Append(srchMenuItem)
        self.Bind(wx.EVT_MENU, self.panel.OnSearch, id=SEARCH_FOOTPRINTS)

        copyMenuItem = wx.MenuItem(srchMenu, COPY_FOOTPRINTS, "Copy\tCtrl+C")
        srchMenu.Append(copyMenuItem)
        self.Bind(wx.EVT_MENU, self.panel.OnCopy, id=COPY_FOOTPRINTS)

        # Help menu containing help and about buttons.
        helpMenu = wx.Menu()
        menuBar.Append(helpMenu, "&Help")

        helpMenuItem = wx.MenuItem(helpMenu, SHOW_HELP, "Help\tCtrl+H")
        helpMenu.Append(helpMenuItem)

        aboutMenuItem = wx.MenuItem(helpMenu, SHOW_ABOUT, "About App\tCtrl+A")
        helpMenu.Append(aboutMenuItem)

        self.Bind(wx.EVT_MENU, self.ShowHelp, id=SHOW_HELP)
        self.Bind(wx.EVT_MENU, self.ShowAbout, id=SHOW_ABOUT)

        self.SetMenuBar(menuBar)

    def OnSearchPath(self, event):
        # Update search path for footprints.

        dlg = TextEntryDialog(
            self,
            title="Set Footprint Search Path",
            caption="Footprint Search Path",
            tip="Enter {sep}-separated list of directories in which to search for fp-lib-table file.".format(
                sep=os.pathsep
            ),
        )
        dlg.Center()
        dlg.SetValue(os.pathsep.join(footprint_search_paths[KICAD]))
        if dlg.ShowModal() == wx.ID_OK:
            footprint_search_paths[KICAD] = dlg.GetValue().split(os.pathsep)
            skidl_cfg.store()  # Stores updated search path in file.
        dlg.Destroy()

    def ShowHelp(self, e):
        Feedback(
            """
1. Enter text to search for in the footprint descriptions.
2. Start the search by pressing Return or clicking on the Search button.
3. Matching footprints will appear in the table in the left-hand pane.
4. Select a row in the table to display footprint info in the right-hand pane.
5. Click the Copy button to place the selected footprint on the clipboard.
6. Paste the clipboard contents into your SKiDL code.
            """,
            "Help",
        )

    def ShowAbout(self, e):
        Feedback(
            APP_TITLE
            + """
(c) 2019 XESS Corp.
https://github.com/xesscorp/skidl
MIT License
            """,
            "About",
        )

    def OnQuit(self, e):
        self.Close()


class FootprintPaintingPanel(wx.Panel):
    def __init__(self, parent):
        super(self.__class__, self).__init__(parent)
        self.footprint = None

        # Repaint the footprint panel whenever the "show actual size" checkbox changes.
        self.paint_actual_size_ckbx = wx.FindWindowById(
            PAINT_ACTUAL_SIZE_CKBX_ID, parent
        )
        self.paint_actual_size_ckbx.Bind(wx.EVT_CHECKBOX, self.OnSize)

        # Repaint the footprint panel for any paint or windows resize event.
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_SIZE, self.OnSize)

    @property
    def footprint(self):
        return self._footprint

    @footprint.setter
    def footprint(self, fp):
        if not fp:
            self._footprint = None
        else:
            self._footprint = FootprintPainter(fp, wx.ClientDC(self))

    def OnPaint(self, event):
        if self.footprint:
            self.Refresh()  # Causes repaint of the entire footprint panel.
            self.footprint.paint(
                wx.PaintDC(self), self.paint_actual_size_ckbx.GetValue()
            )
        else:
            self.ClearBackground()

    def OnSize(self, event):
        if self.footprint:
            self.Refresh()  # Causes repaint of the entire footprint panel.
            self.footprint.paint(
                wx.ClientDC(self), self.paint_actual_size_ckbx.GetValue()
            )
        else:
            self.ClearBackground()


class FootprintSearchPanel(wx.SplitterWindow):
    def __init__(self, *args, **kwargs):
        kwargs["id"] = FOOTPRINT_PANEL_ID
        super(self.__class__, self).__init__(*args, **kwargs)

        # Subpanel for search text box and lib/part table.
        self.search_panel = self.InitSearchPanel(self)

        # Subpanel for part/pin data.
        self.fp_panel = self.InitFootprintPanel(self)

        # Split subpanels left/right.
        self.SplitVertically(
            add_border(self.search_panel, wx.RIGHT),
            add_border(self.fp_panel, wx.LEFT),
            sashPosition=0,
        )
        self.SetSashGravity(0.5)  # Both subpanels expand/contract equally.
        self.SetMinimumPaneSize(MINIMUM_PANE_SIZE)

        # This flag is used to set focus on the table of found footprints
        # after a search is completed.
        self.focus_on_found_footprints = False

        self.Bind(wx.EVT_IDLE, self.OnIdle)

        # Bind event for passing footprint search terms from part to footprint panel.
        self.Bind(EVT_SEND_SEARCH_TERMS, self.OnSearchTerms)

        # Bind event for requesting footprint to be sent from footprint to part panel.
        self.Bind(EVT_REQUEST_FOOTPRINT, self.OnCopy)

        # Using a SplitterWindow shows a corrupted scrollbar area for
        # the default found_footprints table. To eliminate that, draw the table large
        # enough to need a scrollbar, and then draw it at its default size.
        self.found_footprints.Resize(200)  # Draw it large to create scrollbar.
        self.Update()
        self.found_footprints.Resize(10)  # Draw it small to remove scrollbar.
        self.Update()

    def OnIdle(self, event):
        if self.focus_on_found_footprints:
            self.found_footprints.SelectRow(0)
            self.found_footprints.GoToCell(0, 1)
            self.found_footprints.SetFocus()
            self.focus_on_found_footprints = False

    def OnSearchTerms(self, event):
        # Handle data sent from Part panel to Footprint panel.
        self.search_text.Clear()
        self.search_text.WriteText(event.search_terms)

    def InitSearchPanel(self, parent):
        # Subpanel for search text box and footprint table.
        search_panel = wx.Panel(parent)
        vbox = wx.BoxSizer(wx.VERTICAL)
        search_panel.SetSizer(vbox)

        # Text box for footprint search string.
        self.search_text = wx.TextCtrl(search_panel, style=wx.TE_PROCESS_ENTER)
        self.search_text.Bind(wx.EVT_TEXT_ENTER, self.OnSearch)
        tip = wx.ToolTip("Enter text or regular expression to select footprints.")
        self.search_text.SetToolTip(tip)

        # Button to initiate search for footprints containing search string.
        search_btn = wx.Button(search_panel, label="Search")
        search_btn.Bind(wx.EVT_BUTTON, self.OnSearch)
        tip = wx.ToolTip(
            "Search for footprints containing the text or regular expression."
        )
        search_btn.SetToolTip(tip)

        # Table (grid) for holding footprints that match search string.
        self.found_footprints = MyGrid(search_panel, ("Library", "Footprint"))
        self.found_footprints.Resize(10)
        self.found_footprints.Bind(wx.grid.EVT_GRID_SELECT_CELL, self.OnSelectCell)
        self.found_footprints.Bind(wx.grid.EVT_GRID_CELL_LEFT_DCLICK, self.OnCopy)

        # Button to copy selected footprint to clipboard.
        copy_btn = wx.Button(search_panel, label="Copy")
        copy_btn.Bind(wx.EVT_BUTTON, self.OnCopy)
        tip = wx.ToolTip("Copy the selected footprint to the clipboard.")
        copy_btn.SetToolTip(tip)

        # Grid for arranging text box, grid and buttons.
        fgs = wx.FlexGridSizer(rows=2, cols=2, vgap=SPACING, hgap=SPACING)
        fgs.AddMany(
            [
                search_btn,
                (self.search_text, 1, wx.EXPAND),
                copy_btn,
                (self.found_footprints, 1, wx.EXPAND),
            ]
        )
        fgs.AddGrowableCol(1, 1)  # widths of text box and grid are adjustable.
        fgs.AddGrowableRow(1, 1)  # height of grid is adjustable.

        vbox.Add(fgs, proportion=1, flag=wx.ALL | wx.EXPAND, border=0)

        return search_panel

    def InitFootprintPanel(self, parent):
        # Subpanel for footprint painting.
        fp_panel = wx.Panel(parent)
        vbox = wx.BoxSizer(wx.VERTICAL)
        fp_panel.SetSizer(vbox)

        # Text box for displaying description of footprint highlighted in grid.
        self.fp_desc = Description(fp_panel, "Footprint Description/Tags")
        vbox.Add(self.fp_desc, proportion=0, flag=wx.ALL, border=0)
        # Hide the inactive description *after* adding it to the sizer so it's placed correctly.
        self.fp_desc.Hide()

        # Hyperlink for highlighted footprint datasheet.
        self.datasheet_link = HyperLink(fp_panel, label="Datasheet")
        vbox.Add(self.datasheet_link, proportion=0, flag=wx.ALL, border=0)
        # Hide the inactive link *after* adding it to the sizer so it's placed correctly.
        self.datasheet_link.Hide()

        # Footprint painting.
        self.painting_title_panel = wx.Panel(fp_panel)
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        self.painting_title_panel.SetSizer(hbox)
        hbox.Add(
            wx.StaticText(self.painting_title_panel, label="Footprint"),
            proportion=0,
            flag=wx.ALL,
            border=0,
        )
        hbox.AddSpacer(3 * SPACING)
        self.actual_size_ckbx = wx.CheckBox(
            self.painting_title_panel,
            id=PAINT_ACTUAL_SIZE_CKBX_ID,
            label="Show actual size",
        )
        self.actual_size_ckbx.SetToolTip(
            wx.ToolTip("Check to display actual size of footprint.")
        )
        hbox.Add(self.actual_size_ckbx, proportion=0, flag=wx.ALL, border=0)
        vbox.Add(self.painting_title_panel, proportion=0, flag=wx.ALL, border=SPACING)

        self.fp_painting_panel = FootprintPaintingPanel(fp_panel)
        vbox.Add(
            self.fp_painting_panel,
            proportion=1,
            flag=wx.ALL | wx.EXPAND,
            border=SPACING,
        )

        return fp_panel

    def OnSearch(self, event):

        # Setup indicators to show progress while scanning libraries.
        wx.BeginBusyCursor()
        progress = wx.ProgressDialog(
            "Searching Footprint Libraries",
            "Loading footprints from libraries.",
            style=wx.PD_CAN_ABORT | wx.PD_AUTO_HIDE,
        )

        # Scan libraries looking for footprints that match search string.
        self.footprints = []
        search_text = self.search_text.GetLineText(0)
        for lib_module in search_footprints_iter(search_text):
            if lib_module[0] == "LIB":
                lib_name = lib_module[1]
                lib_idx = lib_module[2]
                total_num_libs = lib_module[3]
                progress.SetRange(total_num_libs)
                if not progress.Update(
                    lib_idx, "Reading footprint library {}...".format(lib_name)
                ):
                    # Cancel button was pressed, so abort.
                    progress.Destroy()
                    wx.EndBusyCursor()
                    return
            elif lib_module[0] == "MODULE":
                self.footprints.append(lib_module[1:])

        # Remove progress indicators after search is done.
        progress.Destroy()
        wx.EndBusyCursor()

        # place libraries and parts into a table.
        grid = self.found_footprints

        # Clear any existing grid cells and add/sub rows to hold search results.
        grid.Resize(len(self.footprints))

        # Places libs and part names into table.
        for row, (lib, module_text, module_name) in enumerate(self.footprints):
            grid.SetCellValue(row, 0, lib)
            grid.SetCellValue(row, 1, module_name)

        # Initially sort table by footprint library in ascending order.
        grid.SortTable(0, 1)

        # Size the columns for their new contents.
        grid.AutoSizeColumns()

        # Cause refresh.
        self.search_panel.Layout()

        # Focus on the first footprint in the list.
        self.focus_on_found_footprints = True

    def OnSelectCell(self, event):
        # When a row of the footprint table is selected, display the data for that footprint.

        # Ths is a null footprint that just paints an "X" (cross) when there's no valid footprint.
        null_module_text = [
            "(module NULL",
            "(fp_line (start 0.0 0.0) (end 1.0 1.0) (layer F.Fab) (width 0.01))",
            "(fp_line (start 1.0 0.0) (end 0.0 1.0) (layer F.Fab) (width 0.01))",
            ")",
        ]

        # Get the selected row in the lib/footprint table and translate it to the row in the data table.
        self.found_footprints.ClearSelection()
        self.found_footprints.SelectRow(event.GetRow())
        row = self.found_footprints.GetDataRowIndex(event.GetRow())

        # Get the text describing the footprint structure.
        try:
            module_text = self.footprints[row][1]
        except (AttributeError, IndexError):
            # No module text, so use the null module.
            module_text = null_module_text

        # Parse the footprint text.
        try:
            wx.BeginBusyCursor()
            module = pym.Module.parse("\n".join(module_text))
        except Exception as e:
            # Parsing error, use the null module.
            Feedback(
                "Error while parsing {}:{}".format(
                    self.footprints[row][0], self.footprints[row][2]
                ),
                "Error",
            )
            module_text = null_module_text
            module = pym.Module.parse("\n".join(module_text))
        wx.EndBusyCursor()

        # Get the footprint description and tags if they exist.
        descr = ""
        tags = ""
        for line in module_text:
            try:
                # Get the text following "(descr ".
                descr = line.split("(descr ")[1].rsplit(")", 1)[0]
                descr = rmv_quotes(descr)
            except IndexError:
                pass  # This line didn't have "(descr " in it.
            try:
                # Get the text following "(tags ".
                tags = line.split("(tags ")[1].rsplit(")", 1)[0]
                tags = rmv_quotes(tags)
                tags = "\nTags: " + tags
            except IndexError:
                pass  # This line didn't have "(tags " in it.

        # Show the footprint description.
        self.fp_desc.SetDescription(descr + tags)

        # Display the link to the footprint datasheet.
        try:
            # Extract link-like text from the description.
            url = re.search("(?P<url>https?://[^\s]+)", descr).group("url")
        except AttributeError:
            # No URL found so hide the hyperlink.
            self.datasheet_link.SetURL(None)
        else:
            # URL was found, so set the hyperlink and show it.
            # Often, the link has punctuation at the end that should be removed.
            url = url.rstrip(string.punctuation)
            self.datasheet_link.SetURL(url)

        # Set the footprint that will be displayed.
        self.painting_title_panel.Show()
        self.fp_painting_panel.Show()
        self.fp_painting_panel.footprint = module

        # Re-layout the panel to account for link hide/show.
        self.fp_panel.Layout()

    def OnCopy(self, event):
        # Copy the lib/footprint for the selected footprint onto the clipboard.

        # Get the cell where the cursor is.
        # row = self.found_footprints.GetGridCursorRow()
        try:
            row = self.found_footprints.GetSelectedRows()[0]
        except (IndexError, TypeError):
            return

        # Deselect all rows but the first.
        self.found_footprints.SelectRow(row)

        # Create a SKiDL part instantiation.
        lib = self.found_footprints.GetCellValue(row, 0)
        footprint = self.found_footprints.GetCellValue(row, 1)
        footprint_inst = "footprint='{lib}:{footprint}'".format(**locals())

        # Make a data object to hold the SKiDL part instantiation.
        dataObj = wx.TextDataObject()
        dataObj.SetText(footprint_inst)

        # Place the SKiDL part instantiation on the clipboard.
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(dataObj)
            wx.TheClipboard.Flush()
        else:
            Feedback("Unable to open clipboard!", "Error")

        # Create search string for part footprint.
        evt = SendFootprintEvent(footprint=footprint_inst)
        wx.PostEvent(wx.FindWindowById(PART_PANEL_ID), evt)


def main():

    # import wx.lib.inspection
    app = wx.App()
    AppFrame(None)
    # wx.lib.inspection.InspectionTool().Show()
    app.MainLoop()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2019 by XESS Corp.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
GUI for finding/displaying parts.
"""

from __future__ import print_function

import collections
import os

import wx

from skidl import (
    KICAD,
    lib_search_paths,
    search_parts_iter,
    skidl_cfg,
)
from .common import *

APP_TITLE = "SKiDL Part Search"

APP_EXIT = 1
SHOW_HELP = 3
SHOW_ABOUT = 4
SEARCH_PATH = 5
SEARCH_PARTS = 6
COPY_PART = 7


# Named tuple for parts found by library search.
LibPart = collections.namedtuple("LibPart", "lib_name part part_name")


class AppFrame(wx.Frame):
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

        self.panel = PartSearchPanel(self)
        box = wx.BoxSizer(wx.VERTICAL)
        box.Add(self.panel, proportion=1, flag=wx.ALL | wx.EXPAND, border=SPACING)
        self.SetSizer(box)

        # Keep border same color as background of panel.
        self.SetBackgroundColour(self.panel.GetBackgroundColour())

        self.InitMenus()

        self.SetTitle(APP_TITLE)
        self.Center()
        self.Show(True)
        self.Fit()

    def InitMenus(self):

        # Top menu.
        menuBar = wx.MenuBar()

        # File submenu containing quit button.
        fileMenu = wx.Menu()
        menuBar.Append(fileMenu, "&File")

        quitMenuItem = wx.MenuItem(fileMenu, APP_EXIT, "Quit\tCtrl+Q")
        fileMenu.Append(quitMenuItem)
        self.Bind(wx.EVT_MENU, self.OnQuit, id=APP_EXIT)

        # Search submenu containing search and copy buttons.
        srchMenu = wx.Menu()
        menuBar.Append(srchMenu, "&Search")

        srchPathItem = wx.MenuItem(srchMenu, SEARCH_PATH, "Set search path...\tCtrl+P")
        srchMenu.Append(srchPathItem)
        self.Bind(wx.EVT_MENU, self.OnSearchPath, id=SEARCH_PATH)

        srchMenuItem = wx.MenuItem(srchMenu, SEARCH_PARTS, "Search\tCtrl+F")
        srchMenu.Append(srchMenuItem)
        self.Bind(wx.EVT_MENU, self.panel.OnSearch, id=SEARCH_PARTS)

        copyMenuItem = wx.MenuItem(srchMenu, COPY_PART, "Copy\tCtrl+C")
        srchMenu.Append(copyMenuItem)
        self.Bind(wx.EVT_MENU, self.panel.OnCopy, id=COPY_PART)

        # Help menu containing help and about buttons.
        helpMenu = wx.Menu()
        menuBar.Append(helpMenu, "&Help")

        helpMenuItem = wx.MenuItem(helpMenu, SHOW_HELP, "Help\tCtrl+H")
        helpMenu.Append(helpMenuItem)

        aboutMenuItem = wx.MenuItem(helpMenu, SHOW_ABOUT, "About App\tCtrl+A")
        helpMenu.Append(aboutMenuItem)

        self.Bind(wx.EVT_MENU, self.ShowHelp, id=SHOW_HELP)
        self.Bind(wx.EVT_MENU, self.ShowAbout, id=SHOW_ABOUT)

        self.SetMenuBar(menuBar)

    def OnSearchPath(self, event):
        # Update search path for parts.

        dlg = TextEntryDialog(
            self,
            title="Set Part Search Path",
            caption="Part Search Path",
            tip="Enter {sep}-separated list of directories in which to search for parts.".format(
                sep=os.pathsep
            ),
        )
        dlg.Center()
        dlg.SetValue(os.pathsep.join(lib_search_paths[KICAD]))
        if dlg.ShowModal() == wx.ID_OK:
            lib_search_paths[KICAD] = dlg.GetValue().split(os.pathsep)
            skidl_cfg.store()  # Stores updated lib search path in file.
        dlg.Destroy()

    def ShowHelp(self, e):
        Feedback(
            """
1. Enter text to search for in the part descriptions.
2. Start the search by pressing Return or clicking on the Search button.
3. Matching parts will appear in the Library/Part table in the left-hand pane.
4. Select a row in the Library/Part table to display part info in the right-hand pane.
5. Click the Copy button to place the selected library and part on the clipboard.
6. Paste the clipboard contents into your SKiDL code.
            """,
            "Help",
        )

    def ShowAbout(self, e):
        Feedback(
            APP_TITLE
            + """
(c) 2019 XESS Corp.
https://github.com/xesscorp/skidl
MIT License
            """,
            "About",
        )

    def OnQuit(self, e):
        self.Close()


class PartSearchPanel(wx.SplitterWindow):
    def __init__(self, *args, **kwargs):
        kwargs["id"] = PART_PANEL_ID
        super(self.__class__, self).__init__(*args, **kwargs)

        # Subpanel for search text box and lib/part table.
        self.search_panel = self.InitSearchPanel(self)

        # Subpanel for part/pin data.
        self.part_panel = self.InitPartPanel(self)

        # Split subpanels left/right.
        self.SplitVertically(
            add_border(self.search_panel, wx.RIGHT),
            add_border(self.part_panel, wx.LEFT),
            sashPosition=0,
        )
        self.SetSashGravity(0.5)  # Both subpanels expand/contract equally.
        self.SetMinimumPaneSize(MINIMUM_PANE_SIZE)

        # This flag is used to set focus on the table of found parts
        # after a search is completed.
        self.focus_on_found_parts = False

        self.Bind(wx.EVT_IDLE, self.OnIdle)

        # Bind event for passing data from footprint panel to part panel.
        self.Bind(EVT_SEND_FOOTPRINT, self.OnFootprint)

        # Using a SplitterWindow shows a corrupted scrollbar area for
        # the default found_parts table. To eliminate that, draw the table large
        # enough to need a scrollbar, and then draw it at its default size.
        self.found_parts.Resize(200)  # Draw it large to create scrollbar.
        self.Update()
        self.found_parts.Resize(10)  # Draw it small to remove scrollbar.
        self.Update()

    def OnIdle(self, EnvironmentError):
        # Perform some delayed actions when the event processor is idle.

        if self.focus_on_found_parts:
            # Once a search is done, place the cursor on the first entry
            # in the new table of found parts.
            self.found_parts.SelectRow(0)
            self.found_parts.GoToCell(0, 1)
            self.found_parts.SetFocus()
            self.focus_on_found_parts = False

    def InitSearchPanel(self, parent):
        # Subpanel for search text box and lib/part table.
        search_panel = wx.Panel(parent)
        vbox = wx.BoxSizer(wx.VERTICAL)
        search_panel.SetSizer(vbox)

        # Text box for part search string.
        self.search_text = wx.TextCtrl(search_panel, style=wx.TE_PROCESS_ENTER)
        self.search_text.Bind(wx.EVT_TEXT_ENTER, self.OnSearch)
        tip = wx.ToolTip("Enter text or regular expression to select parts.")
        self.search_text.SetToolTip(tip)

        # Button to initiate search for parts containing search string.
        search_btn = wx.Button(search_panel, label="Search")
        search_btn.Bind(wx.EVT_BUTTON, self.OnSearch)
        tip = wx.ToolTip("Search for parts containing the text or regular expression.")
        search_btn.SetToolTip(tip)

        # Table (grid) for holding libs and parts that match search string.
        self.found_parts = MyGrid(search_panel, ("Library", "Part"))
        self.found_parts.Resize(10)
        self.found_parts.Bind(wx.grid.EVT_GRID_SELECT_CELL, self.OnSelectCell)
        self.found_parts.Bind(wx.grid.EVT_GRID_CELL_LEFT_DCLICK, self.OnCopy)

        # Button to copy selected lib/part to clipboard.
        copy_btn = wx.Button(search_panel, label="Copy")
        copy_btn.Bind(wx.EVT_BUTTON, self.OnCopy)
        tip = wx.ToolTip("Copy the selected part to the clipboard.")
        copy_btn.SetToolTip(tip)

        # Checkbox for enabling creation of a TEMPLATE-type part.
        self.template_ckbx = wx.CheckBox(
            search_panel,
            # id=TEMPLATE_CKBX_ID,
            label="Tmplt",
        )
        tip = wx.ToolTip("Check to create a part TEMPLATE instead of a part INSTANCE.")
        self.template_ckbx.SetToolTip(tip)

        # Grid for arranging text box, grid and buttons.
        fgs = wx.FlexGridSizer(rows=2, cols=2, vgap=SPACING, hgap=SPACING)
        fgs.AddMany(
            [
                search_btn,
                (self.search_text, 1, wx.EXPAND),
                stack_it(
                    copy_btn, self.template_ckbx
                ),  # Copy & template buttons stacked together.
                (self.found_parts, 1, wx.EXPAND),
            ]
        )
        fgs.AddGrowableCol(1, 1)  # widths of text box and grid are adjustable.
        fgs.AddGrowableRow(1, 1)  # height of grid is adjustable.

        vbox.Add(fgs, proportion=1, flag=wx.ALL | wx.EXPAND, border=0)

        return search_panel

    def InitPartPanel(self, parent):
        # Subpanel for part/pin data.
        part_panel = wx.Panel(parent)
        vbox = wx.BoxSizer(wx.VERTICAL)
        part_panel.SetSizer(vbox)

        # Text box for displaying description of part highlighted in grid.
        self.part_desc = Description(part_panel, "Part Description/Tags")
        vbox.Add(self.part_desc, proportion=0, flag=wx.ALL, border=0)
        # Hide the inactive description *after* adding it to the sizer so it's placed correctly.
        self.part_desc.Hide()

        # Hyperlink for highlighted part datasheet.
        self.datasheet_link = HyperLink(part_panel, label="Datasheet")
        vbox.Add(self.datasheet_link, proportion=0, flag=wx.ALL, border=0)
        # Hide the inactive link *after* adding it to the sizer so it's placed correctly.
        self.datasheet_link.Hide()

        # Table (grid) of part pin numbers, names, I/O types.
        vbox.Add(
            wx.StaticText(part_panel, label="Pin List"),
            proportion=0,
            flag=wx.ALL,
            border=SPACING,
        )
        self.pin_info = MyGrid(part_panel, ("Pin", "Name", "Type", "Unit"))
        self.pin_info.SetSelectionMode(wx.grid.Grid.GridSelectionModes.GridSelectCells)
        self.pin_info.Resize(10)
        self.pin_info.SetSortFunc(0, natural_sort_key)  # Natural sort pin numbers.
        self.pin_info.SetSortFunc(1, natural_sort_key)  # Natural sort pin names.
        self.pin_info.SetSortFunc(3, natural_sort_key)  # Natural sort list of units.
        vbox.Add(self.pin_info, proportion=1, flag=wx.ALL | wx.EXPAND, border=SPACING)

        return part_panel

    def OnSearch(self, event):

        # Setup indicators to show progress while scanning libraries.
        wx.BeginBusyCursor()
        progress = wx.ProgressDialog(
            "Searching Part Libraries",
            "Loading parts from libraries.",
            style=wx.PD_CAN_ABORT | wx.PD_AUTO_HIDE,
        )

        # Scan libraries looking for parts that match search string.
        self.lib_parts = set()
        search_text = self.search_text.GetLineText(0)
        for lib_part in search_parts_iter(search_text):
            if lib_part[0] == "LIB":
                lib_name, lib_idx, num_libs = lib_part[1:4]
                progress.SetRange(num_libs)
                if not progress.Update(
                    lib_idx, "Reading library {}...".format(lib_name)
                ):
                    # Cancel button was pressed, so abort.
                    progress.Destroy()
                    wx.EndBusyCursor()
                    return
            elif lib_part[0] == "PART":
                self.lib_parts.add(LibPart(*lib_part[1:]))

        # Remove progress indicators after search is done.
        progress.Destroy()
        wx.EndBusyCursor()

        # Sort parts by libraries and part names.
        self.lib_parts = sorted(
            list(self.lib_parts), key=lambda x: "/".join([x.lib_name, x.part_name])
        )

        # place libraries and parts into a table.
        grid = self.found_parts

        # Clear any existing grid cells and add/sub rows to hold search results.
        grid.Resize(len(self.lib_parts))

        # Places libs and part names into table.
        for row, lib_part in enumerate(self.lib_parts):
            grid.SetCellValue(row, 0, lib_part.lib_name)
            grid.SetCellValue(row, 1, lib_part.part_name)

        # Initially sort table by part library in ascending order.
        grid.SortTable(0, 1)

        # Size the columns for their new contents.
        grid.AutoSizeColumns()

        # Focus on the first part in the list.
        self.focus_on_found_parts = True

    def OnSelectCell(self, event):
        # When a row of the lib/part table is selected, display the data for that part.

        # Get the selected row in the lib/part table and translate it to the row in the data table.
        row = self.found_parts.GetDataRowIndex(event.GetRow())

        # Fully instantiate the selected part.
        try:
            part = self.lib_parts[row].part
        except (AttributeError, IndexError):
            return  # Nothing in the lib_parts table.

        part.parse()  # Instantiate pins.

        # Show the part description.
        desc = part.description
        # if part.aliases:
        #     desc += "\nAliases: " + ", ".join(list(part.aliases))
        if part.keywords:
            desc += "\nKeywords: " + part.keywords
        self.part_desc.SetDescription(desc)

        # Display the link to the part datasheet.
        self.datasheet_link.SetURL(part.datasheet)
        if part.datasheet and part.datasheet not in ("~",):
            self.datasheet_link.SetURL(part.datasheet)
            self.datasheet_link.Show()
        else:
            self.datasheet_link.Hide()
        # Re-layout the panel to account for link hide/show.
        self.part_panel.Layout()

        # Place pin data into a table.
        grid = self.pin_info

        # Clear any existing pin data and add/sub rows to hold results.
        grid.Resize(len(part))

        # Sort pins by pin number.
        pins = sorted(part, key=lambda p: natural_sort_key(p.get_pin_info()[0]))

        # Place pin data into the table.
        unit_pins = dict()
        for lbl, u in part.unit.items():
            unit_pins[lbl] = [p.num for p in u.pins]
        for row, pin in enumerate(pins):
            num, names, func = pin.get_pin_info()
            units = []
            for lbl in part.unit:
                if num in unit_pins[lbl]:
                    units.append(lbl)
            grid.SetCellValue(row, 0, num)
            grid.SetCellValue(row, 1, names)
            grid.SetCellValue(row, 2, func)
            grid.SetCellValue(row, 3, ",".join(units))

        # Size the columns for their new contents.
        grid.AutoSizeColumns()

        # Create search string for part footprint.
        # First, search for any of the footprints in the fplist of the part.
        fp_srch_terms = "|".join(part.fplist)
        fp_srch_terms = re.sub("[*?]", ".\g<0>", fp_srch_terms)
        # Use the number of pins in the part to refine the search.
        num_pins = len(part.pins)
        fp_srch_terms = " ".join((fp_srch_terms, "^#pads={}$".format(num_pins)))

        # Send the string of search terms to the footprint search panel.
        evt = SendSearchTermsEvent(search_terms=fp_srch_terms)
        wx.PostEvent(wx.FindWindowById(FOOTPRINT_PANEL_ID), evt)

    def OnCopy(self, event):
        # Copy the lib/part for the selected part onto the clipboard.

        # Get the cell where the cursor is.
        row = self.found_parts.GetGridCursorRow()

        # Deselect all rows but the first.
        self.found_parts.SelectRow(row)

        # Create a SKiDL part instantiation without a footprint.
        lib = self.found_parts.GetCellValue(row, 0)
        part = self.found_parts.GetCellValue(row, 1)
        if self.template_ckbx.IsChecked():
            s = "Part(lib='{lib}', name='{part}', dest=TEMPLATE)"
        else:
            s = "Part(lib='{lib}', name='{part}')"
        self.part_inst = s.format(**locals())

        # Make a data object to hold the SKiDL part instantiation.
        dataObj = wx.TextDataObject()
        dataObj.SetText(self.part_inst)

        # Place the SKiDL part instantiation on the clipboard.
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(dataObj)
            wx.TheClipboard.Flush()
        else:
            Feedback("Unable to open clipboard!", "Error")

        # Request the footprint panel to return whatever footprint is selected.
        # The footprint panel will respond with an event containing the footprint.
        evt = RequestFootprintEvent()
        wx.PostEvent(wx.FindWindowById(FOOTPRINT_PANEL_ID), evt)

    def OnFootprint(self, evt):
        # Receive the selected footprint from the footprint panel and
        # place a lib/part with a footprint onto the clipboard.

        # Remove the end parenthesis of the non-footprint part instance
        # and append the footprint selected in the footprint panel.
        try:
            part_inst = "{part_inst}, {footprint})".format(
                part_inst=self.part_inst[:-1], footprint=evt.footprint
            )
        except AttributeError:
            return  # No part was copied.

        # Make a data object to hold the SKiDL part instantiation.
        dataObj = wx.TextDataObject()
        dataObj.SetText(part_inst)

        # Place the SKiDL part instantiation on the clipboard.
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(dataObj)
            wx.TheClipboard.Flush()
        else:
            Feedback("Unable to open clipboard!", "Error")


def main():

    # import wx.lib.inspection
    app = wx.App()
    AppFrame(None)
    # wx.lib.inspection.InspectionTool().Show()
    app.MainLoop()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2019 by XESS Corp.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


"""
GUI components in common with multiple apps.
"""

import re

import wx
import wx.grid
import wx.lib.agw.hyperlink as hl
import wx.lib.expando
import wx.lib.newevent

MINIMUM_PANE_SIZE = 300
SPACING = 5

# IDs for part and footprint panels.
PART_PANEL_ID = wx.NewId()
FOOTPRINT_PANEL_ID = wx.NewId()

# Events for part and footprint panels to cooperate with each other.
# Request footprint panel to send selected footprint.
RequestFootprintEvent, EVT_REQUEST_FOOTPRINT = wx.lib.newevent.NewEvent()
# Send selected footprint to part panel.
SendFootprintEvent, EVT_SEND_FOOTPRINT = wx.lib.newevent.NewEvent()
# Send search terms from part panel to footprint panel.
SendSearchTermsEvent, EVT_SEND_SEARCH_TERMS = wx.lib.newevent.NewEvent()


def natural_sort_key(s, _nsre=re.compile("([0-9]+)")):
    """Sorting function for pin numbers or names."""
    return [int(text) if text.isdigit() else text.lower() for text in _nsre.split(s)]


def is_dark_mode():
    window_colour = wx.SystemSettings.GetColour(wx.SYS_COLOUR_WINDOW)
    luminance = sum([c * f for c, f in zip(window_colour, [0.299, 0.587, 0.114])])
    return luminance < 128


def box_it(window, title_text):
    panel = wx.Panel(window.GetParent())
    window.Reparent(panel)
    sbox = wx.StaticBox(panel, label=title_text)
    box = wx.StaticBoxSizer(sbox, wx.VERTICAL)
    box.Add(window, proportion=1, flag=wx.ALL | wx.EXPAND, border=0)
    panel.SetSizer(box)
    return panel


def stack_it(*windows):
    """Stack windows vertically."""

    stack = wx.Panel(windows[0].GetParent())
    for window in windows:
        window.Reparent(stack)

    sizer = wx.BoxSizer(wx.VERTICAL)
    for window in windows[:-1]:
        sizer.Add(window, proportion=1, flag=wx.BOTTOM | wx.EXPAND, border=SPACING)
    sizer.Add(windows[-1], proportion=1, flag=wx.ALL | wx.EXPAND, border=0)

    stack.SetSizer(sizer)

    return stack


def add_title(window, title_text, location):
    """Add title to top/bottom of a window."""

    titled_window = wx.Panel(window.GetParent())
    window.Reparent(titled_window)

    title = wx.StaticText(
        titled_window, label=title_text, style=wx.ALIGN_CENTRE_HORIZONTAL
    )

    if is_dark_mode():
        title.SetBackgroundColour(wx.Colour(128, 128, 128))
        title.SetForegroundColour(wx.Colour(55, 55, 55))
    else:
        title.SetBackgroundColour(wx.Colour(128, 128, 128))
        title.SetForegroundColour(wx.Colour(200, 200, 200))

    title.SetFont(title.GetFont().MakeLarger().MakeBold().MakeItalic())
    box = wx.BoxSizer(wx.VERTICAL)

    if location == wx.TOP:
        box.Add(title, proportion=0, flag=wx.BOTTOM | wx.EXPAND, border=SPACING)
        box.Add(window, proportion=1, flag=wx.ALL | wx.EXPAND, border=0)
    else:
        box.Add(window, proportion=1, flag=wx.ALL | wx.EXPAND, border=0)
        box.Add(title, proportion=0, flag=wx.TOP | wx.EXPAND, border=SPACING)

    titled_window.SetSizer(box)

    return titled_window


def add_border(window, location):
    """Add border line to one side of a window."""

    bordered_window = wx.Panel(window.GetParent())
    window.Reparent(bordered_window)

    if location in (wx.TOP, wx.BOTTOM):
        border = wx.StaticLine(bordered_window, size=(-1, 2))
        box = wx.BoxSizer(wx.VERTICAL)
    else:
        border = wx.StaticLine(bordered_window, size=(2, -1))
        box = wx.BoxSizer(wx.HORIZONTAL)

    tip = wx.ToolTip("Drag sash to resize panes.")
    tip.SetDelay(0)
    border.SetToolTip(tip)

    if location == wx.TOP:
        box.Add(border, proportion=0, flag=wx.BOTTOM | wx.EXPAND, border=SPACING)
        box.Add(window, proportion=1, flag=wx.ALL | wx.EXPAND, border=0)
    elif location == wx.LEFT:
        box.Add(border, proportion=0, flag=wx.RIGHT | wx.EXPAND, border=SPACING)
        box.Add(window, proportion=1, flag=wx.ALL | wx.EXPAND, border=0)
    elif location == wx.BOTTOM:
        box.Add(window, proportion=1, flag=wx.ALL | wx.EXPAND, border=0)
        box.Add(border, proportion=0, flag=wx.TOP | wx.EXPAND, border=SPACING)
    else:
        box.Add(window, proportion=1, flag=wx.ALL | wx.EXPAND, border=0)
        box.Add(border, proportion=0, flag=wx.LEFT | wx.EXPAND, border=SPACING)

    bordered_window.SetSizer(box)

    return bordered_window


def Feedback(msg, label):
    """Show a dialog with a message and an OK button."""

    dlg = wx.MessageDialog(None, msg, label, wx.OK)
    dlg.ShowModal()
    dlg.Destroy()


class Description(wx.Panel):
    """Class for showing a text description in a TextCtrl box."""

    def __init__(self, parent, label):
        super(self.__class__, self).__init__(parent)

        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)

        vbox.Add(
            wx.StaticText(self, label=label),
            proportion=0,
            flag=wx.ALL,
            border=SPACING / 2,
        )

        self.desc = wx.TextCtrl(
            self, size=(10000, 60), style=wx.TE_READONLY | wx.TE_MULTILINE
        )
        vbox.Add(self.desc, proportion=0, flag=wx.ALL, border=SPACING)

        vbox.Add(
            wx.StaticLine(self, size=(10000, 2), style=wx.LI_HORIZONTAL),
            proportion=0,
            flag=wx.ALL | wx.EXPAND,
            border=SPACING / 2,
        )

    def SetDescription(self, description):
        """Set the description shown in the text box."""

        self.desc.Remove(0, self.desc.GetLastPosition())
        if not description:
            self.Hide()
        else:
            self.desc.WriteText(description)
            self.Show()


class HyperLink(wx.Panel):
    """Class for showing a clickable hyperlink."""

    def __init__(self, parent, label):
        super(self.__class__, self).__init__(parent)

        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)

        self.link = hl.HyperLinkCtrl(self, label=label, URL="")
        self.link.EnableRollover(True)
        vbox.Add(self.link, proportion=0, flag=wx.ALL, border=SPACING)

        vbox.Add(
            wx.StaticLine(self, size=(10000, 2), style=wx.LI_HORIZONTAL),
            proportion=0,
            flag=wx.ALL | wx.EXPAND,
            border=SPACING / 2,
        )

    def SetURL(self, url):
        """Set the URL for the clickable hyperlink."""

        self.link.SetURL(url)
        if not url:
            self.Hide()
        else:
            self.Show()


class TextEntryDialog(wx.Dialog):
    """Class for entering text in a dialog window."""

    def __init__(self, parent, title, caption, tip=None):

        # Create dialog window.
        style = wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER
        super(self.__class__, self).__init__(parent, -1, title, style=style)

        # Create text label describing the purpose of the window.
        text = wx.StaticText(self, -1, caption)

        # Create text entry box.
        self.input = wx.lib.expando.ExpandoTextCtrl(
            self, size=(int(0.75 * parent.GetSize()[0]), -1), style=wx.TE_PROCESS_ENTER
        )
        self.input.Bind(wx.EVT_TEXT_ENTER, self.OnEnter)

        # Add tool tip to text entry box.
        if tip:
            self.input.SetToolTip(wx.ToolTip(tip))

        # Add OK and Cancel buttons.
        buttons = self.CreateButtonSizer(wx.OK | wx.CANCEL)

        # Arrange all the items in the window.
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(text, 0, wx.ALL, 5)
        sizer.Add(self.input, 1, wx.EXPAND | wx.ALL, 5)
        sizer.Add(buttons, 0, wx.EXPAND | wx.ALL, 5)
        self.SetSizerAndFit(sizer)

    def OnEnter(self, event):
        """End modal if enter is hit in input text control."""
        self.EndModal(wx.ID_OK)

    def SetValue(self, value):
        """Set the initial value shown in the text box."""
        self.input.SetValue(value.strip().rstrip())
        self.input.SetFocus()
        self.input.SetInsertionPointEnd()
        self.Fit()

    def GetValue(self):
        """Get the value entered in the text box."""
        return self.input.GetValue().strip().rstrip()


class MyGrid(wx.grid.Grid):
    """Class for displaying tabular data about parts, pins, footprints."""

    DEFAULT_NUM_ROWS = 10
    SPACER = "  "  # Spacer for begin/end of column labels.
    ASCENDING = " ▲"  # Indicator that column is sorted in ascending order.
    DESCENDING = " ▼"  # Indicator that column is sorted in descending order.

    def __init__(self, parent, headers):

        # Create grid with a column for each header label.
        super(self.__class__, self).__init__()
        self.Create(parent)
        self.CreateGrid(
            numRows=self.DEFAULT_NUM_ROWS,
            numCols=len(headers),
            selmode=wx.grid.Grid.SelectRows,
        )

        # Set table attributes.
        if is_dark_mode():
            # Set cell background and cursor colors for dark mode.
            self.BackgroundColour = wx.Colour(20, 20, 20)
            self.SetCellHighlightColour(wx.Colour(235, 235, 235))
        else:
            # Set cell background and cursor colors for light mode.
            self.BackgroundColour = wx.Colour(255, 255, 255)
            self.SetCellHighlightColour(wx.Colour(20, 20, 20))
        self.HideRowLabels()  # Hide row labels 1, 2, 3...
        self.EnableEditing(False)  # User can't edit values in table.
        self.SetDefaultCellBackgroundColour(parent.GetBackgroundColour())
        self.ColourGridBackground()
        self.SetSelectionMode(wx.grid.Grid.GridSelectionModes.GridSelectRows)
        self.SetLabelFont(self.GetLabelFont().MakeBold())
        self.SetTabBehaviour(wx.grid.Grid.Tab_Leave)

        # Set column header labels.
        for col, lbl in enumerate(headers):
            self.SetColLabelValue(col, self.SPACER + lbl + self.SPACER)

        # Set the initial column to sort by and the sorting direction.
        self.sorting = {"col": 0, "dir": 0}

        # Set sorting function for each column.
        self.sort_funcs = [lambda x: x] * len(headers)

        # Set event that triggers sorting of the table rows.
        self.Bind(wx.grid.EVT_GRID_LABEL_LEFT_DCLICK, self.OnSort)

        # Set event that copies selected cells to the clipboard.
        self.Bind(wx.grid.EVT_GRID_CELL_RIGHT_CLICK, self.OnCopy)

    def Resize(self, numRows=0):
        """Clear the table and resize the number of rows."""

        if numRows == 0:
            numRows = self.DEFAULT_NUM_ROWS
        self.ClearGrid()
        num_rows_chg = numRows - self.GetNumberRows()
        if num_rows_chg < 0:
            self.DeleteRows(0, -num_rows_chg, True)
        elif num_rows_chg > 0:
            self.AppendRows(num_rows_chg)
        self.SetSortingIndicator(0, 0)  # No sorting at the start.
        self.ColourGridBackground()

        # Create a list of row indices that will be sorted along with the
        # table. Then when data moves to a different row, the index associated
        # with that row will indicate the index into the original data set.
        self.data_row_indices = list(range(numRows))

    def ColourGridBackground(self):
        """Set the background color of all the cells in the table."""

        for r in range(self.GetNumberRows()):
            for c in range(self.GetNumberCols()):
                self.SetCellBackgroundColour(r, c, self.BackgroundColour)

    def SetSortingIndicator(self, new_col, new_dir):
        """Set the sorting indicator for a given table column."""

        # Remove sorting indicator from old column label.
        old_col = self.sorting["col"]
        self.SetColLabelValue(
            old_col, self.GetColLabelValue(old_col)[: -len(self.SPACER)] + self.SPACER
        )

        # Set the sorting indicator for the new column.
        if new_dir < 0:
            indicator = self.DESCENDING
        elif new_dir == 0:
            indicator = self.SPACER
        else:
            indicator = self.ASCENDING
        self.SetColLabelValue(
            new_col, self.GetColLabelValue(new_col)[: -len(indicator)] + indicator
        )
        self.sorting["dir"] = new_dir
        self.sorting["col"] = new_col

    def SetSortFunc(self, col, func):
        """Set the sorting function for a particular table column."""

        self.sort_funcs[col] = func

    def SortTable(self, sort_col, sort_dir):
        """Sort the table rows based on the values in a particular column."""

        self.SetSortingIndicator(sort_col, sort_dir)
        n_rows = self.GetNumberRows()
        n_cols = self.GetNumberCols()

        # Copy rows of data from the cells of the table into a sortable list.
        tbl_vals = []
        for row in range(n_rows):
            row_vals = []
            for col in range(n_cols):
                row_vals.append(self.GetCellValue(row, col))
            # Append the index of the original row location to each row of data.
            # This is used to find the original data values after the rows have
            # been moved during sorting.
            row_vals.append(self.data_row_indices[row])
            tbl_vals.append(row_vals)  # Append row of values to list.

        # This function will extract the value from the sorting column
        # of a row of data, apply the sorting function for that column to
        # the data, and return the result to the Python sort routine.
        def sort_func(data_row):
            return self.sort_funcs[sort_col](data_row[sort_col])

        # Sort the list of data rows based on the values in one of the columns.
        tbl_vals.sort(key=sort_func, reverse=(sort_dir < 0))

        # Re-enter the data into the cells of the table.
        for row, row_vals in enumerate(tbl_vals):
            for col, col_val in enumerate(row_vals[:-1]):
                self.SetCellValue(row, col, col_val)
            # Update the indices for where the data in each row of the table
            # came from in the original data set.
            self.data_row_indices[row] = row_vals[-1]

    def GetDataRowIndex(self, tbl_row):
        """
        For a given row in the table of cells, return the index of the row
        in the original set of data used to create the table.
        """
        return self.data_row_indices[tbl_row]

    def OnSort(self, event):
        """Sort the table based on which column header is clicked."""

        sort_col = event.GetCol()
        if sort_col != self.sorting["col"]:
            # If a new column was selected for sorting, always start in ascending mode.
            sort_dir = 1
        else:
            # If the same sorting column was selected, then toggle the sorting direction.
            sort_dir = -self.sorting["dir"] or 1  # If dir==0, then set it to 1.

        # Sor the data in the table of cells.
        self.SortTable(sort_col, sort_dir)

    def OnCopy(self, event):
        num_rows = self.GetNumberRows()
        num_cols = self.GetNumberCols()
        cells = []
        for cell in self.GetSelectedCells():
            cells.append((cell[0], cell[1]))
        for selrow in self.GetSelectedRows():
            for c in range(num_cols):
                cells.append((selrow, c))
        for selcol in self.GetSelectedCols():
            for r in range(num_rows):
                cells.append((r, selcol))
        for topleft, bottomright in zip(
            self.GetSelectionBlockTopLeft(), self.GetSelectionBlockBottomRight()
        ):
            for r in range(topleft[0], bottomright[0] + 1):
                for c in range(topleft[1], bottomright[1] + 1):
                    cells.append((r, c))

        cells.sort()

        prev_row = cells[0][0]  # First row in cell coords.
        vals = ""
        for row, col in cells:
            if row != prev_row:
                vals += "\n"
            vals += '"' + str(self.GetCellValue(row, col)) + '"' + ", "
            prev_row = row

        # Make a data object to hold the cell values.
        dataObj = wx.TextDataObject()
        dataObj.SetText(vals[:-2])  # Strip off last ", "

        # Place the cell values on the clipboard.
        if wx.TheClipboard.Open():
            wx.TheClipboard.SetData(dataObj)
            wx.TheClipboard.Flush()
        else:
            Feedback("Unable to open clipboard!", "Error")



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2019 by XESS Corp.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
GUI for finding/displaying parts and footprints.
"""

from __future__ import print_function

import os

import wx

from skidl import (
    KICAD,
    SchLib,
    footprint_cache,
    footprint_search_paths,
    lib_search_paths,
    skidl_cfg,
)
from .common import *
from .pckg_info import __version__
from .skidl_footprint_search import FootprintSearchPanel
from .skidl_part_search import PartSearchPanel

APP_TITLE = "zyc: SKiDL Part/Footprint Search"
APP_EXIT = 1
SHOW_HELP = 3
SHOW_ABOUT = 4
PART_SEARCH_PATH = 5
FOOTPRINT_SEARCH_PATH = 6
REFRESH = 7


class AppFrame(wx.Frame):
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

        self.panel = PartFootprintSearchPanel(self)
        box = wx.BoxSizer(wx.VERTICAL)
        box.Add(self.panel, proportion=1, flag=wx.ALL | wx.EXPAND, border=SPACING)
        self.SetSizer(box)

        # Keep border same color as background of panel.
        self.SetBackgroundColour(self.panel.GetBackgroundColour())

        self.InitMenus()

        self.SetTitle(APP_TITLE)
        self.Center()
        self.Show(True)
        self.Fit()

    def InitMenus(self):

        # Top menu.
        menuBar = wx.MenuBar()

        # File submenu containing quit button.
        fileMenu = wx.Menu()
        menuBar.Append(fileMenu, "&File")

        quitMenuItem = wx.MenuItem(fileMenu, APP_EXIT, "Quit\tCtrl+Q")
        fileMenu.Append(quitMenuItem)
        self.Bind(wx.EVT_MENU, self.OnQuit, id=APP_EXIT)

        # Search submenu containing search and copy buttons.
        srchMenu = wx.Menu()
        menuBar.Append(srchMenu, "&Search")

        partSrchPathItem = wx.MenuItem(
            srchMenu, PART_SEARCH_PATH, "Set part search path...\tCtrl+P"
        )
        srchMenu.Append(partSrchPathItem)
        self.Bind(wx.EVT_MENU, self.OnPartSearchPath, id=PART_SEARCH_PATH)

        footprintSrchPathItem = wx.MenuItem(
            srchMenu, FOOTPRINT_SEARCH_PATH, "Set footprint search path...\tCtrl+F"
        )
        srchMenu.Append(footprintSrchPathItem)
        self.Bind(wx.EVT_MENU, self.OnFootprintSearchPath, id=FOOTPRINT_SEARCH_PATH)

        refreshItem = wx.MenuItem(srchMenu, REFRESH, "Refresh part + footprint paths")
        srchMenu.Append(refreshItem)
        self.Bind(wx.EVT_MENU, self.OnRefresh, id=REFRESH)

        # Help menu containing help and about buttons.
        helpMenu = wx.Menu()
        menuBar.Append(helpMenu, "&Help")

        helpMenuItem = wx.MenuItem(helpMenu, SHOW_HELP, "Help\tCtrl+H")
        helpMenu.Append(helpMenuItem)

        aboutMenuItem = wx.MenuItem(helpMenu, SHOW_ABOUT, "About App\tCtrl+A")
        helpMenu.Append(aboutMenuItem)

        self.Bind(wx.EVT_MENU, self.ShowHelp, id=SHOW_HELP)
        self.Bind(wx.EVT_MENU, self.ShowAbout, id=SHOW_ABOUT)

        self.SetMenuBar(menuBar)

    def OnPartSearchPath(self, event):
        # Update search path for parts.

        dlg = TextEntryDialog(
            self,
            title="Set Part Search Path",
            caption="Part Search Path",
            tip="Enter {sep}-separated list of directories in which to search for parts.".format(
                sep=os.pathsep
            ),
        )
        dlg.Center()
        dlg.SetValue(os.pathsep.join(lib_search_paths[KICAD]))
        if dlg.ShowModal() == wx.ID_OK:
            lib_search_paths[KICAD] = dlg.GetValue().split(os.pathsep)
            skidl_cfg.store()  # Stores updated lib search path in file.
        dlg.Destroy()

    def OnFootprintSearchPath(self, event):
        # Update search path for footprints.

        dlg = TextEntryDialog(
            self,
            title="Set Footprint Search Path",
            caption="Footprint Search Path",
            tip="Enter {sep}-separated list of directories in which to search for fp-lib-table file.".format(
                sep=os.pathsep
            ),
        )
        dlg.Center()
        dlg.SetValue(os.pathsep.join(footprint_search_paths[KICAD]))
        if dlg.ShowModal() == wx.ID_OK:
            footprint_search_paths[KICAD] = dlg.GetValue().split(os.pathsep)
            skidl_cfg.store()  # Stores updated search path in file.
        dlg.Destroy()

    def OnRefresh(self, event):
        SchLib.reset()
        footprint_cache.reset()

    def ShowHelp(self, e):
        Feedback(
            """
1. Enter keywords/regex in the part search box.
2. Matching parts will appear in the Library/Part table.
3. Select a row in the Library/Part table to display part info.
4. Enter keywords/regex in the footprint search box.
5. Matching footprints will appear in the Library/Footprint table.
6. Select a row in the Library/Footprint table to display the footprint.
7. a) Click the Copy button in the Part Search panel to copy 
      the part & footprint to the clipboard, -OR-
   b) Click the Copy button in the Footprint Search panel to copy
      the footprint to the clipboard, -OR-
   c) Deselect (ctrl-click) the footprint row and click the
      Copy button in the Part Search panel to copy just
      the part to the clipboard.
8. Paste the clipboard contents into your SKiDL code.

General:
* Drag sashes to resize individual panels.
* Double-click column headers to sort table contents.
* Ctrl-click to select/deselect table cells.
            """,
            "Help",
        )

    def ShowAbout(self, e):
        Feedback(
            APP_TITLE + " " + __version__
            + """
(c) 2019 XESS Corp.
https://github.com/xesscorp/skidl
MIT License
            """,
            "About",
        )

    def OnQuit(self, e):
        self.Close()


class PartFootprintSearchPanel(wx.SplitterWindow):
    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

        # Subpanel for part search panel.
        self.part_panel = add_border(
            add_title(PartSearchPanel(self), "Part Search", wx.TOP), wx.BOTTOM
        )
        # self.part_panel = box_it(PartSearchPanel(self), "Part Search")

        # Subpanel for footprint search.
        self.footprint_panel = add_border(
            add_title(FootprintSearchPanel(self), "Footprint Search", wx.TOP), wx.TOP
        )
        # self.footprint_panel = box_it(FootprintSearchPanel(self), "Footprint Search")

        # Split subpanels top/bottom.
        self.SplitHorizontally(self.part_panel, self.footprint_panel, sashPosition=0)
        self.SetSashGravity(0.5)  # Both subpanels expand/contract equally.
        self.Update()


def main():
    # import wx.lib.inspection

    app = wx.App()
    AppFrame(None)
    # wx.lib.inspection.InspectionTool().Show()
    app.MainLoop()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# MIT license
#
# Copyright (C) 2019 by XESS Corp.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Class for painting a PCB footprint.
"""

from __future__ import print_function

import math
from collections import defaultdict, namedtuple

import wx

# Prescaling constant increases footprint dimensions
# so they won't lose precision when converted to integers
# during painting operations.
PRESCALE = 100000

MIN_SIZE = 1e-7

# KiCad layer colors.
FP_BCK_COLOUR = wx.Colour(0, 0, 0)
FCU_COLOUR = wx.Colour(128, 0, 0)
BCU_COLOUR = wx.Colour(20, 81, 132)
FPASTE_COLOUR = wx.Colour(132, 0, 0)
BPASTE_COLOUR = wx.Colour(0, 194, 194)
FSILK_COLOUR = wx.Colour(0, 132, 132)
BSILK_COLOUR = wx.Colour(132, 0, 132)
FMASK_COLOUR = wx.Colour(132, 0, 132)
BMASK_COLOUR = wx.Colour(132, 132, 0)
FCRTYD_COLOUR = wx.Colour(132, 132, 132)
BCRTYD_COLOUR = wx.Colour(194, 194, 0)
FFAB_COLOUR = wx.Colour(194, 194, 0)
BFAB_COLOUR = wx.Colour(132, 0, 0)
DRILL_COLOUR = FP_BCK_COLOUR  # Same as background color so drills look like holes.

# KiCad layer brushes.
FCU_BRUSH = wx.BRUSHSTYLE_SOLID
BCU_BRUSH = wx.BRUSHSTYLE_SOLID
FFAB_BRUSH = wx.BRUSHSTYLE_SOLID
BFAB_BRUSH = wx.BRUSHSTYLE_SOLID
FCRTYD_BRUSH = wx.BRUSHSTYLE_SOLID
BCRTYD_BRUSH = wx.BRUSHSTYLE_SOLID
FSILK_BRUSH = wx.BRUSHSTYLE_SOLID
BSILK_BRUSH = wx.BRUSHSTYLE_SOLID
DRILL_BRUSH = wx.BRUSHSTYLE_SOLID

# Color & brush for each layer.
layer_style = {
    "B.Cu": (BCU_COLOUR, BCU_BRUSH),
    "F.Cu": (FCU_COLOUR, FCU_BRUSH),
    "F&B.Cu": (FCU_COLOUR, FCU_BRUSH),
    "*.Cu": (FCU_COLOUR, FCU_BRUSH),
    "B.Fab": (BFAB_COLOUR, BFAB_BRUSH),
    "F.Fab": (FFAB_COLOUR, FFAB_BRUSH),
    "F&B.Fab": (FFAB_COLOUR, FFAB_BRUSH),
    "*.Fab": (FFAB_COLOUR, FFAB_BRUSH),
    "B.CrtYd": (BCRTYD_COLOUR, BCRTYD_BRUSH),
    "F.CrtYd": (FCRTYD_COLOUR, FCRTYD_BRUSH),
    "F&B.CrtYd": (FCRTYD_COLOUR, FCRTYD_BRUSH),
    "*.CrtYd": (FCRTYD_COLOUR, FCRTYD_BRUSH),
    "B.SilkS": (BSILK_COLOUR, BSILK_BRUSH),
    "F.SilkS": (FSILK_COLOUR, FSILK_BRUSH),
    "F&B.SilkS": (FSILK_COLOUR, FSILK_BRUSH),
    "*.SilkS": (FSILK_COLOUR, FSILK_BRUSH),
    "Drill": (DRILL_COLOUR, DRILL_BRUSH),
}

# Named tuple for storing a bounding box.
BBox = namedtuple("BBox", "x0 y0 x1 y1")


class Layer(object):
    """Stores graphic primitives for a layer and draws them."""

    def __init__(self, colour=wx.Colour(63, 63, 63), style=wx.BRUSHSTYLE_SOLID):
        self.set_fill(colour, style)
        self.clear()  # Creates all the empty graphic element lists.

    def set_fill(self, colour, style=wx.BRUSHSTYLE_SOLID):
        """Set layer color and fill style."""

        self.colour = colour
        if isinstance(colour, wx.Colour):
            self.brush = wx.Brush(colour, style)
        elif isinstance(colour, wx.Bitmap):
            self.brush = wx.Brush(colour)
        else:
            raise NotImplementedError

    def clear(self):
        """Clear all graphic primitives."""

        self.pad_polygons = []
        self.pad_circles = []
        self.lines = []
        self.circles = []

    def add_pad_polygon(self, polygon):
        """Add list of polygon points to the layer."""
        self.pad_polygons.append(polygon)

    def add_pad_circle(self, circle):
        """Add center & radius of a pad disk to the layer."""
        self.pad_circles.append(circle)

    def add_line(self, line):
        """Add start & end points for a line segment to the layer."""
        self.lines.append(line)

    def add_circle(self, circle):
        """Add center & radius of a circle to the layer."""
        self.circles.append(circle)

    def paint(self, dc):
        """Paint the graphic primitives into the device context."""

        # For filled polygons, use a transparent pen and the layer brush style.
        dc.SetPen(wx.Pen(wx.Colour(0, 0, 0), style=wx.PENSTYLE_TRANSPARENT))
        dc.SetBrush(self.brush)
        dc.DrawPolygonList(self.pad_polygons)

        # For pad circles, use a transparent pen and the layer brush style.
        for pad_circle in self.pad_circles:
            circ = list(pad_circle)[:]
            circ[2] = max(circ[2], 2)  # no circle gets painted with radius<2.
            dc.DrawCircle(*circ)

        # For lines, use the layer color.
        pen = wx.Pen(self.colour)
        for line in self.lines:
            pen.SetWidth(max(line[0], 1))  # Set pen width.
            dc.SetPen(pen)
            dc.DrawLine(*line[1:])

        # For circles, use the layer color.
        for circle in self.circles:
            pen.SetWidth(max(circle[0], 1))  # Set pen width.
            dc.SetPen(pen)
            circ = list(circle[1:])  # Get circle center and radius.
            circ[2] = max(
                circ[2], 2
            )  # Radius must be greater than 2 or no circle is drawn.
            dc.DrawCircle(*circ)


class FootprintPainter(object):
    """Footprint painted within a device context."""

    def __init__(self, footprint_, dc):
        self.footprint = footprint_  # Set footprint data.
        self.bbox = None  # Clear the bbox so it will be calculated.
        self.paint(dc)  # This will calculate the footprint bbox.

    @staticmethod
    def rect_pts(cx, cy, w, h, rot, tm):
        """Compute the scaled/translated/rotated endpoints for a rectangle."""

        cx, cy, w, h = [m * PRESCALE for m in [cx, cy, w, h]]
        h2, w2 = h / 2, w / 2
        pts = ((-w2, -h2), (-w2, h2), (w2, h2), (w2, -h2))
        tmp_tm = wx.AffineMatrix2D(tm)
        tmp_tm.Translate(cx, cy)
        tmp_tm.Rotate(rot * math.pi / 180)
        return [list(tmp_tm.TransformPoint(pt)) for pt in pts]

    @staticmethod
    def trapezoid_pts(cx, cy, w, h, dw, dh, rot, tm):
        """Compute the scaled/translated/rotated endpoints for a trapezoid."""

        cx, cy, w, h, dw, dh = [m * PRESCALE for m in [cx, cy, w, h, dw, dh]]
        h2, w2, dw2, dh2 = h / 2, w / 2, dw / 2, dh / 2
        pts = (
            (-w2 + dw2, -h2 - dh2),
            (-w2 - dw2, h2 + dh2),
            (w2 + dw2, h2 - dh2),
            (w2 - dw2, -h2 + dh2),
        )
        tmp_tm = wx.AffineMatrix2D(tm)
        tmp_tm.Translate(cx, cy)
        tmp_tm.Rotate(rot * math.pi / 180)
        return [list(tmp_tm.TransformPoint(pt)) for pt in pts]

    @staticmethod
    def circle_pts(cx, cy, r, tm):
        """Compute the scaled/translated/rotated center & radius for a circle."""

        cx, cy, r = [m * PRESCALE for m in [cx, cy, r]]
        cx, cy = tm.TransformPoint(cx, cy)
        r, _ = tm.TransformDistance(r, 0)
        return (cx, cy, r)

    @staticmethod
    def line_pts(w, x0, y0, x1, y1, tm):
        """Compute the scaled/translated/rotated endpoints for a line segment."""

        w, x0, y0, x1, y1 = [m * PRESCALE for m in [w, x0, y0, x1, y1]]
        w, _ = tm.TransformDistance(w, 0)
        x0, y0 = tm.TransformPoint(x0, y0)
        x1, y1 = tm.TransformPoint(x1, y1)
        return (w, x0, y0, x1, y1)

    @staticmethod
    def roundrect_pts(cx, cy, w, h, r, rot, tm):
        """Compute the scaled/translated/rotated endpoints for a rounded rectangle."""

        # Build a rounded rectangle from two overlapping rectangles with
        # four circles to cover the corners.

        # Overlap two rectangles to create a larger rectangle with the corners cut out.
        rects = []
        rects.append(FootprintPainter.rect_pts(cx, cy, w, h - 2 * r, rot, tm))
        rects.append(FootprintPainter.rect_pts(cx, cy, w - 2 * r, h, rot, tm))

        # Add four circles to fill-in corners.
        circles = []
        # Matrix for rotating circle centers and translating by pad center.
        tmp_tm = wx.AffineMatrix2D()
        tmp_tm.Translate(cx, cy)
        tmp_tm.Rotate(rot * math.pi / 180)
        # Circle at NE corner cutout.
        ccx, ccy = tmp_tm.TransformPoint(w / 2 - r, h / 2 - r)
        circles.append(FootprintPainter.circle_pts(ccx, ccy, r, tm))
        # Circle at NW corner cutout.
        ccx, ccy = tmp_tm.TransformPoint(-w / 2 + r, h / 2 - r)
        circles.append(FootprintPainter.circle_pts(ccx, ccy, r, tm))
        # Circle at SE corner cutout.
        ccx, ccy = tmp_tm.TransformPoint(w / 2 - r, -h / 2 + r)
        circles.append(FootprintPainter.circle_pts(ccx, ccy, r, tm))
        # Circle at SW corner cutout.
        ccx, ccy = tmp_tm.TransformPoint(-w / 2 + r, -h / 2 + r)
        circles.append(FootprintPainter.circle_pts(ccx, ccy, r, tm))

        return rects, circles

    @staticmethod
    def mk_bmp(w, h, colour, pctg):
        """Make a bitmap pattern."""

        sz = len(colour)
        n = w * h * sz
        buff = bytearray(n)
        acc = 0
        for i in range(0, n, sz):
            acc += pctg
            if acc >= 1:
                buff[i : i + sz] = colour
                acc -= 1
        if sz == 4:
            bmp = wx.Bitmap.FromBufferRGBA(w, h, buff)
        elif sz == 3:
            bmp = wx.Bitmap.FromBuffer(w, h, buff)
        else:
            raise NotImplementedError
        bmp_mask = wx.Mask(bmp, wx.Colour(0, 0, 0))
        bmp.SetMask(bmp_mask)
        return bmp

    def paint(self, dc, paint_actual_size=False):
        """Paint the footprint into the device context."""

        bbox = self.bbox
        if bbox:
            # The bbox for the footprint exists, so set the
            # scaling and translation for painting it within the DC.
            panel_w, panel_h = dc.GetSize()
            if paint_actual_size:
                scale_x = dc.GetPPI().GetWidth() * 1.0 / (25.4 * PRESCALE)
                scale_y = dc.GetPPI().GetHeight() * 1.0 / (25.4 * PRESCALE)
            else:
                panel_w = panel_w or 100  # Prevent division by zero.
                panel_h = panel_h or 100  # Prevent division by zero.
                scale_x = float(panel_w) / (bbox.x1 - bbox.x0)
                scale_y = float(panel_h) / (bbox.y1 - bbox.y0)
            scale = min(scale_x, scale_y)

            # Compute translation to place footprint center at center of panel.
            bbox_cx, bbox_cy = (bbox.x1 + bbox.x0) / 2, (bbox.y1 + bbox.y0) / 2
            panel_cx, panel_cy = (panel_w / 2) / scale, (panel_h / 2) / scale
            tx, ty = panel_cx - bbox_cx, panel_cy - bbox_cy

            # Transformation matrices (TMs) operate opposite of expected:
            # the first scale/trans operation you add is the last operation
            # that's applied to a point.
            tm = wx.AffineMatrix2D()  # Start with identity matrix.
            tm.Scale(scale, scale)  # Apply scaling from real to screen dimensions.
            tm.Translate(tx, ty)
        else:
            # No bbox, so this must be first time the footprint is being painted.
            # Therefore, paint it without scaling to determine the physical dimensions.
            # These will be used later to scale it to the panel display.
            tm = wx.AffineMatrix2D()  # Identity matrix so no scaling.

        # Initialize device context for drawing.
        dc.SetBackground(wx.Brush(FP_BCK_COLOUR))
        dc.Clear()

        # Create layer dictionary for storing graphics on each layer.
        layers = defaultdict(Layer)

        # Draw the footprint's pad primitives.
        for pad in self.footprint.pads:
            attr = pad.attributes

            # Process the pad.
            w, h = (attr["size"] + [0])[0:2]
            cx, cy, rot = (attr["at"] + [0])[0:3]
            rot = -rot

            # Apply the offset of the pad from the drill.
            try:
                offset_x, offset_y = attr["drill"].attributes["offset"]
            except (AttributeError, KeyError, TypeError):
                pass  # Either no drill or no pad offset w.r.t. the drill.
            else:
                # The pad is offset w.r.t. the drill, so it goes in the opposite dir.
                offset_x, offset_y = -offset_x, -offset_y
                # Rotate the offset as needed.
                rot_tm = wx.AffineMatrix2D()
                rot_tm.Rotate(rot * math.pi / 180)
                offset_x, offset_y = rot_tm.TransformPoint(offset_x, offset_y)
                # Apply the offset to the center of the pad.
                cx -= offset_x
                cy -= offset_y

            # Paint the appropriate pad shape.
            if attr["shape"] == "rect":
                pts = self.rect_pts(cx, cy, w, h, rot, tm)
                for lyr_nm in attr["layers"]:
                    layers[lyr_nm].add_pad_polygon(pts)
            elif attr["shape"] == "circle":
                r = w / 2
                pts = self.circle_pts(cx, cy, r, tm)
                for lyr_nm in attr["layers"]:
                    layers[lyr_nm].add_pad_circle(pts)
            elif attr["shape"] == "roundrect":
                r_ratio = attr["roundrect_rratio"]
                r = r_ratio * min(w, h)
                pts_sets = self.roundrect_pts(cx, cy, w, h, r, rot, tm)
                for lyr_nm in attr["layers"]:
                    for rect_pts in pts_sets[0]:
                        layers[lyr_nm].add_pad_polygon(rect_pts)
                    for circle_pts in pts_sets[1]:
                        layers[lyr_nm].add_pad_circle(circle_pts)
            elif attr["shape"] == "oval":
                r_ratio = 0.5
                r = r_ratio * min(w, h)
                pts_sets = self.roundrect_pts(cx, cy, w, h, r, rot, tm)
                for lyr_nm in attr["layers"]:
                    for rect_pts in pts_sets[0]:
                        layers[lyr_nm].add_pad_polygon(rect_pts)
                    for circle_pts in pts_sets[1]:
                        layers[lyr_nm].add_pad_circle(circle_pts)
            elif attr["shape"] == "trapezoid":
                try:
                    dh, dw = attr["rect_delta"]
                except TypeError:
                    dh, dw = 0, 0
                pts = self.trapezoid_pts(cx, cy, w, h, dw, dh, rot, tm)
                for lyr_nm in attr["layers"]:
                    layers[lyr_nm].add_pad_polygon(pts)
            else:
                raise NotImplementedError

            # Process any drill in the pad.
            cx, cy, rot = (attr["at"] + [0])[0:3]
            try:
                # See if there's a drill in this pad.
                drill_attr = attr["drill"].attributes
            except (KeyError, AttributeError):
                pass  # No drill for this pad.
            else:
                # Yes, there is a drill.
                lyr_nm = "Drill"
                size = drill_attr["size"]
                try:
                    # See if this pad is oval (i.e., has w, h).
                    w, h = size
                    r_ratio = 0.5
                    r = r_ratio * min(w, h)
                    pts_sets = self.roundrect_pts(cx, cy, w, h, r, rot, tm)
                    for rect_pts in pts_sets[0]:
                        layers[lyr_nm].add_pad_polygon(rect_pts)
                    for circle_pts in pts_sets[1]:
                        layers[lyr_nm].add_pad_circle(circle_pts)
                except TypeError:
                    # No, pad is circular.
                    r = size / 2
                    pts = self.circle_pts(cx, cy, r, tm)
                    layers[lyr_nm].add_pad_circle(pts)

        # Draw the footprint's line primitives.
        for line in self.footprint.lines:
            attr = line.attributes
            w = attr["width"]
            x0, y0 = attr["start"]
            x1, y1 = attr["end"]
            pts = self.line_pts(w, x0, y0, x1, y1, tm)
            lyr_nm = attr["layer"]
            layers[lyr_nm].add_line(pts)

        # Draw the footprint's circle primitives.
        for circle in self.footprint.circles:
            attr = circle.attributes
            w = attr["width"]
            cx, cy = attr["center"]
            x1, y1 = attr["end"]
            # Use line_pts routine to calc coords of center and endpoint and line width.
            w, cx, cy, x1, y1 = self.line_pts(w, cx, cy, x1, y1, tm)
            r = math.hypot(cx - x1, cy - y1)
            lyr_nm = attr["layer"]
            layers[lyr_nm].add_circle((w, cx, cy, r))

        # Draw each layer, starting from the bottom layers that will be overwritten by the top layers.
        for lyr_nm in (
            "B.SilkS",
            "B.Fab",
            "B.CrtYd",
            "B.Cu",
            "F&B.Cu",
            "F.Cu",
            "*.Cu",
            "F&B.CrtYd",
            "F.CrtYd",
            "*.CrtYd",
            "F&B.Fab",
            "F.Fab",
            "*.Fab",
            "F&B.SilkS",
            "F.SilkS",
            "*.SilkS",
            "Drill",
        ):
            layers[lyr_nm].set_fill(*layer_style[lyr_nm])  # Set layer color and fill.
            layers[lyr_nm].paint(dc)  # Draw items for that layer.

        # layers['F.Mask'].set_fill(FMASK_COLOUR, wx.BRUSHSTYLE_CROSS_HATCH)
        # bmp = self.mk_bmp(4,4,(0,255,0), 1.0/16)
        # layers['F.Paste'].set_fill(bmp)

        if not self.bbox:
            # Calculate the bounding box if it hasn't been set, yet.
            # In this case, the transformation matrix will be an identity matrix,
            # so the bbox will contain the physical size of the footprint.
            # This will be scaled for the given device context when this function
            # is called again.
            self.bbox = BBox(*dc.GetBoundingBox())

            # make the bbox a little bigger to put some margin along each side.
            delta = 0.025 * max(
                self.bbox.x1 - self.bbox.x0, self.bbox.y1 - self.bbox.y0
            )

            # Correct point-sized bounding boxes to prevent divide-by-zero errors during scaling.
            if delta < MIN_SIZE:
                delta = MIN_SIZE

            self.bbox = BBox(
                self.bbox.x0 - delta,
                self.bbox.y0 - delta,
                self.bbox.x1 + delta,
                self.bbox.y1 + delta,
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

import sys
from setuptools import setup, find_packages

__author__ = "XESS Corp."
__email__ = "info@xess.com"
__version__ = "0.2.0"

if "sdist" in sys.argv[1:]:
    with open("zyc/pckg_info.py", "w") as f:
        for name in ["__version__", "__author__", "__email__"]:
            f.write("{} = '{}'\n".format(name, locals()[name]))

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = [
    "skidl >= 0.0.27",
    "kinparse >= 0.1.0",
    'enum34; python_version < "3.0"',
    "wxpython >= 4.0.7",
    "pykicad",
]

setup_requirements = []

test_requirements = []

setup(
    author=__author__,
    author_email=__email__,
    version=__version__,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Manufacturing",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="A GUI for searching and selecting parts and footprints for use in SKiDL.",
    entry_points={"gui_scripts": ["zyc = zyc.zyc:main"]},
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="zyc",
    name="zyc",
    packages=find_packages(include=["zyc"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/xesscorp/zyc",
    zip_safe=False,
)
